
import re

from copy import deepcopy
from collections import OrderedDict
from collections import Counter
from collections import namedtuple

from ..util.filesystem import write_file
from ..util.filesystem import read_file
from ..util.filesystem import apply_file_extension

from ..util.iter import OrderedDefaultDict
from ..util.iter import iteration_depth
from ..util.iter import modify_iteration_depth
from ..util.iter import transpose
from ..util.iter import map_numeric_indices
from ..util.iter import ordered_unique
from ..util.iter import base_class_names
from ..util.iter import is_vengeance_class
from ..util.iter import is_flux_row_class
from ..util.iter import is_exhaustable
from ..util.iter import is_subscriptable

from ..util.text import p_json_dumps
from ..util.text import print_runtime

from .flux_row_cls import flux_row_cls


class flux_cls:
    """ primary data management class
    (I need to publish my vengeance example project)

    eg:
        flux = flux_cls([['col_a', 'col_b', 'col_c'],
                         ['a',     'b',     'c'],
                         [1,        2,       3],
                         [None,     None,    None]])
    """

    def __init__(self, matrix=None):
        (self.headers,
         self.matrix) = self.to_flux_rows(matrix)

    @classmethod
    def to_flux_rows(cls, matrix, headers=None):
        matrix = cls.__validate_modify_matrix(matrix)

        if headers is None:
            headers = cls.to_flux_row_headers(matrix[0])
        elif not isinstance(headers, dict):
            raise TypeError('headers must be a dictionary')

        matrix = [flux_row_cls(headers, row) for row in matrix]

        return headers, matrix

    @classmethod
    def to_flux_row_headers(cls, names):
        cls.__validate_no_reserved_headers(names)
        return map_numeric_indices(names)

    # region {serialization methods}
    def to_csv(self, path, encoding=None):
        m = self.rows()
        write_file(path, m, encoding=encoding)

    @classmethod
    def from_csv(cls, path, encoding=None):
        m = read_file(path, encoding=encoding)
        return cls(m)

    def to_json(self, path=None, encoding=None):
        j = [row.dict() for row in self.matrix[1:]]
        j = p_json_dumps(j)
        if path is None:
            return j

        write_file(path, j, encoding=encoding)

    @classmethod
    def from_json(cls, path, encoding=None):
        rows = read_file(path, encoding=encoding)
        if not rows:
            m = [[]]
        else:
            m = [list(rows[0].keys())] + \
                [list(d.values()) for d in rows]

        return cls(m)

    def serialize(self, path):
        """
        although convenient, pickle introduces significant security flaws
        you should be sure no malicious actors have access to the location of these files
        """
        path = apply_file_extension(path, '.flux')
        write_file(path, self)

    @classmethod
    def deserialize(cls, path):
        """
        although convenient, pickle introduces significant security flaws
        you should be sure no malicious actors have access to the location of these files
        """
        path = apply_file_extension(path, '.flux')
        return read_file(path)
    # endregion

    @property
    def header_values(self):
        """
        there may be a discrepancy between self.matrix[0].values and self.headers.keys
            map_numeric_indices() makes certain modifications to self.headers.keys,
            such as coercing values to strings, modifying duplicate values, etc
        """
        return list(self.headers.keys())

    @property
    def first_five_rows(self):
        return [row.values for row in self.matrix[:6]]

    @property
    def num_rows(self):
        """ header row excluded """
        return len(self.matrix) - 1

    @property
    def num_cols(self):
        return len(self.matrix[0].values)

    @property
    def max_num_cols(self):
        # m = [row.values for row in self.matrix]
        # return max(map(len, m))

        return max(map(len, self.matrix))

    @property
    def is_jagged(self):
        num_cols = self.num_cols

        for row in self.matrix[1:]:
            if len(row.values) != num_cols:
                return True

        return False

    @property
    def is_empty(self):
        for row in self.matrix:
            if row.values:
                return False

        return True

    def rows(self, r_1=0, r_2=None):
        r_1 = self.__matrix_row_index(r_1)
        r_2 = self.__matrix_row_index(r_2) + 1

        return (row.values for row in self.matrix[r_1:r_2])

    def flux_rows(self, r_1=0, r_2=None):
        r_1 = self.__matrix_row_index(r_1)
        r_2 = self.__matrix_row_index(r_2) + 1

        return (row for row in self.matrix[r_1:r_2])

    def __matrix_row_index(self, r):
        anchors = {'*h': 0,
                   '*f': 1,
                   '*l': self.num_rows,
                   None: self.num_rows}

        r_i = anchors.get(r, r)
        if not isinstance(r_i, int):
            raise ValueError("invalid row reference '{}'".format(r))

        return r_i

    def execute_commands(self, commands, profile=False):
        profiler = self.__profiler_function(profile)
        commands = self.__parse_commands(commands)

        completed_commands = []

        for command in commands:
            completed_commands.append(command.name)

            if profiler:
                method = profiler(command.method)
            else:
                method = command.method

            if command.num_args == 0:
                method()
            elif command.num_args == 1:
                method(command.args)
            else:
                method(*command.args)

        if hasattr(profiler, 'print_stats'):
            profiler.print_stats()

    @staticmethod
    def __profiler_function(profile):
        if not profile:
            return None

        if profile == 'print_runtime':
            return print_runtime

        try:
            # noinspection PyUnresolvedReferences
            from line_profiler import LineProfiler
            return LineProfiler()

        except ImportError as e:
            if str(profile).lower() in ('line_profiler', 'lineprofiler'):
                raise ImportError('line_profiler site-package not installed') from e

        return print_runtime

    def __parse_commands(self, commands):
        parsed = []
        nt_cls = namedtuple('command_nt', ('name',
                                           'method',
                                           'args',
                                           'num_args'))
        for command in commands:
            if isinstance(command, (list, tuple)):
                name, args = command

                if isinstance(args, str):
                    num_args = 1
                else:
                    num_args = len(args)
            else:
                name = command
                args = None
                num_args = 0

            method = getattr(self, name)
            parsed.append(nt_cls(name, method, args, num_args))

        return parsed

    def matrix_by_headers(self, *names):
        names   = modify_iteration_depth(names, depth=1)
        columns = transpose([row.values for row in self.matrix[1:]]) or [[]]

        m = []
        for header in names:
            header = self.__renamed_or_inserted_header(header)

            if header in self.headers:
                i = self.headers[header]
                column = columns[i]
            else:
                column = [None] * self.num_rows

            column.insert(0, header)
            m.append(column)

        m = transpose(m)

        (self.headers,
         self.matrix) = self.to_flux_rows(m)

    def __renamed_or_inserted_header(self, header):
        """
        parentheses around inserted headers:
            to ensure new columns are being created intentionally and not because
            of spelling errors, inserted headers must be surrounded by parenthesis,
            eg: '(inserted_header)'
        """
        if not isinstance(header, (dict, tuple, str)):
            raise ValueError('{}\nheader must be either dictionary or string'.format(header))

        if isinstance(header, dict):
            h_old, h_new = tuple(header.items())[0]
            self.headers[h_new] = self.headers[h_old]

            return h_new

        if isinstance(header, tuple):
            h_old, h_new = header
            self.headers[h_new] = self.headers[h_old]

            return h_new

        if header in self.headers:
            return header

        if header.startswith('(') and header.endswith(')'):
            return header[1:-1]

        if header not in self.headers:
            raise ValueError("'{header}' does not exist\ninserted columns should be surrounded "
                             "by parenthesis, ie '({header})' not '{header}'".format(header=header))

        return header

    def columns(self, *names, mapped=False):
        names = modify_iteration_depth(names, depth=1)
        indices = self.__names_to_column_indices(names, self.headers)

        f = self.__row_values_accessor(indices)
        c = [f(row) for row in self.matrix[1:]]

        if len(indices) > 1:
            c = transpose(c)

        if mapped:
            names = [self.header_values[i] for i in indices]
            if len(names) == 1:
                c = [c]

            c = OrderedDict((n, column) for n, column in zip(names, c))

        return c

    def rename_columns(self, old_to_new_headers):
        header_values = self.matrix[0].values

        for h_old, h_new in old_to_new_headers.items():
            i = self.headers[h_old]
            header_values[i] = h_new

        self._reapply_header_names(header_values)

    def insert_columns(self, *names):
        """ eg:
            flux.insert_columns((0, 'inserted'))         # insert column at beginning
            flux.insert_columns((3, 'inserted'))         # insert column before column 3
            flux.insert_columns(('col_c', 'inserted'))   # insert column before column 'col_c'
            flux.insert_columns((-1, 'inserted'))        # insert column before end
            flux.insert_columns((1, 'inserted_a'),       # insert columns
                                (1, 'inserted_b'),
                                (1, 'inserted_c'))
        """
        if not names:
            return

        names = self.__validate_modify_inserted_names(names, self.headers)
        names = list(reversed(names))

        header_values = self.matrix[0].values

        num_inserted = 0
        for before, header in names:
            if isinstance(before, int):
                i = before
            else:
                i = header_values.index(before)

            header_values.insert(i, header)
            num_inserted += 1

        indices = [header_values.index(h) for _, h in names]
        indices.sort()

        self._reapply_header_names(header_values)

        for i in indices:
            for row in self.matrix[1:]:
                row.values.insert(i, None)

    def append_columns(self, *names, values=None):
        if not names:
            return

        names = modify_iteration_depth(names, depth=1)
        self.__validate_no_duplicates(names)
        self.__validate_no_overlap_with_headers(names, self.headers)

        header_values = self.matrix[0].values
        header_values.extend(names)
        self._reapply_header_names(header_values)

        if values is None:
            columns = [None for _ in range(len(names))]
            for row in self.matrix[1:]:
                row.values.extend(columns)

            return

        if iteration_depth(values) != 1:
            raise ValueError('iteration depth of values must be exactly one')

        if len(values) != self.num_rows:
            raise ValueError('number of values must match the number of rows in flux')

        for row, v in zip(self, values):
            row.values.append(v)

    def delete_columns(self, *names):
        if not names:
            return

        names = modify_iteration_depth(names, depth=1)

        self.__validate_no_duplicates(names)
        self.__validate_no_index_errors_with_headers(names, self.headers)

        indices = self.__names_to_column_indices(names, self.headers)
        indices.sort(reverse=True)

        for row in self.matrix:
            for c in indices:
                del row.values[c]

        self._reapply_header_names(self.matrix[0].values)

    def identify_jagged_rows(self):
        max_cols = self.max_num_cols

        jagged_items = []
        for i, row in enumerate(self.matrix):
            if len(row.values) != max_cols:
                jagged_items.append((i, row))

        return jagged_items

    def insert_rows(self, i, rows):
        if self.is_empty:
            (self.headers,
             self.matrix) = self.to_flux_rows(rows)

            return

        if i == 0:
            del self.matrix[0]

            headers, matrix = self.to_flux_rows(rows)
            self.headers = headers
            self.matrix[i:i] = matrix

            return

        _, matrix = self.to_flux_rows(rows, self.headers)
        self.matrix[i:i] = matrix

    def append_rows(self, rows):
        if self.is_empty:
            (self.headers,
             self.matrix) = self.to_flux_rows(rows)

            return

        _, matrix = self.to_flux_rows(rows, self.headers)
        self.matrix += matrix

    def sort(self, *names, reverse=None):
        """ in-place sort

        eg:
            flux.sort('col_a')
            flux.sort('col_a', 'col_b', 'col_c',
                      reverse=[True, False, True])
        """
        self.matrix[1:] = self.__sort_rows(names, reverse)

    def sorted(self, *names, reverse=None):
        """ :return: sorted flux_cls

        eg:
            flux = flux.sorted('col_a')
            flux = flux.sorted('col_a', 'col_b', 'col_c',
                               reverse=[False, True, True])
        """
        m = [self.matrix[0].values.copy()]
        m += [row.values.copy() for row in self.__sort_rows(names, reverse)]

        flux = self.__class__(m)

        return flux

    def __sort_rows(self, names, order):
        if isinstance(order, bool):
            order = [order]
        else:
            order = order or []

        if not isinstance(order, (list, tuple)):
            raise TypeError('reverse must be a list of booleans')

        names = modify_iteration_depth(names, depth=1)
        for _ in range(len(names) - len(order)):
            order.append(False)

        m = self.matrix[1:]

        # sort priority from left to right; leftmost column sorted last
        names = reversed(names)
        order = reversed(order)

        for f, o in zip(names, order):
            f = self.__row_values_accessor(f)
            o = bool(o)
            m.sort(key=f, reverse=o)

        return m

    def filter(self, f, *args, **kwargs):
        """ in-place """
        self.matrix[1:] = [row for row in self.matrix[1:] if f(row, *args, **kwargs)]

    def filtered(self, f, *args, **kwargs):
        """ :return: new flux_cls """
        m = [self.matrix[0].values.copy()]
        m.extend([row.values.copy() for row in self.matrix[1:] if f(row, *args, **kwargs)])

        flux = self.__class__(m)

        return flux

    def filter_by_unique(self, *f):
        self.__filter_unique_rows(*f, in_place=True)

    def filtered_by_unique(self, *f):
        return self.__filter_unique_rows(*f, in_place=False)

    def __filter_unique_rows(self, *f, in_place):
        def evaluate_unique(row):
            """ first-class function """
            v = f(row)
            if v not in u:
                u.add(v)
                return True
            else:
                return False

        u = set()
        f = self.__row_values_accessor(f)

        if in_place:
            self.filter(evaluate_unique)
        else:
            return self.filtered(evaluate_unique)

    def unique_values(self, *f):
        """ :return: list of ordered, unique values from [f(row)] """
        f = self.__row_values_accessor(f)
        return ordered_unique(f(row) for row in self.matrix[1:])

    def index_row(self, *f):
        """ :return: dictionary of {f(row): row}

        rows with non-unique keys are overwritten
        """
        f = self.__row_values_accessor(f)
        return OrderedDict((f(row), row) for row in self.matrix[1:])

    def index_rows(self, *f):
        """ :return: dictionary of {f(row): [rows]}

        rows with non-unique keys are appended to a list
        """
        f = self.__row_values_accessor(f)
        d = OrderedDefaultDict(list, [(f(row), row) for row in self.matrix[1:]])

        return OrderedDict(d)

    def namedtuples(self):
        self.__validate_headers_as_namedtuple_fields(self.headers)
        flux_row_ntc = namedtuple('flux_row_ntc', self.headers.keys())

        for i, row in enumerate(self.matrix[1:], 1):
            try:
                yield flux_row_ntc(*row.values)
            except TypeError as e:
                raise TypeError('jagged row at index {}?'.format(i)) from e

    def label_row_indices(self, start=0):
        """ meant to assist with debugging;

        label each flux_row_cls.__dict__ with an index, which will then appear
        in each row's __repr__ function and make them easier to identify after
        filtering, sorting, etc
        """
        if 'i' in self.headers:
            raise AssertionError("'i' already exists as header name")

        for i, row in enumerate(self.matrix, start):
            row.__dict__['i'] = i

    def copy(self, deep=False):
        if deep:
            return deepcopy(self)

        m = [row.values.copy() for row in self.matrix]
        flux = self.__class__(m)

        return flux

    def contiguous_indices(self, *f):
        """ :return: yield (i_1, i_2) indices where values are identical """
        if self.is_empty:
            return [(None, None)]

        f = self.__row_values_accessor(f)
        rows   = iter(self)
        v_prev = f(next(rows))

        i_1 = 1
        for i_2, row in enumerate(rows, i_1 + 1):
            v = f(row)

            if v != v_prev:
                yield i_1, i_2 - 1
                v_prev = v
                i_1 = i_2

        yield i_1, self.num_rows

    def replace_matrix(self, m):
        (self.headers,
         self.matrix) = self.to_flux_rows(m)

    def _reapply_header_names(self, headers):
        """
        reassignment of self.headers will de-reference all flux_row_cls._headers
        in self.matrix. To prevent this, self.headers.clear() must be called,
        followed by individual reassignment of new values
        """
        self.headers.clear()
        for h, i in map_numeric_indices(headers).items():
            self.headers[h] = i

    def __row_values_accessor(self, names):
        """ :return: a function that retrieves column values for each flux_row_cls """

        def row_values(row):
            """ return muliple column values from each row """
            return tuple(row.values[_i_] for _i_ in indices)

        def row_value(row):
            """ return single column value from each row """
            return row.values[i]

        names = modify_iteration_depth(names, depth=0)
        if callable(names):
            return names

        self.__validate_no_index_errors_with_headers(modify_iteration_depth(names, 1),
                                                     self.headers)
        if isinstance(names, (list, tuple)):
            indices = self.__names_to_column_indices(names, self.headers)
            f = row_values
        else:
            i = self.headers.get(str(names), names)
            f = row_value

        return f

    def __len__(self):
        return len(self.matrix)

    def __getitem__(self, name):
        """ :return column values """
        if isinstance(name, slice):
            indices = list(range(self.num_cols))
            name = indices[name]

        return self.columns(name)

    def __setitem__(self, name, values):
        """ sets values to a single column

        number of values must be equal to number of rows in matrix,
        excluding header row
        eg:
            flux['col'] = ['blah'] * flux.num_rows
        """
        if iteration_depth(name) > 0:
            raise AssertionError('values must be set one column at a time')

        if iteration_depth(values) == 2:
            values = transpose(values)[0]

        if len(values) != self.num_rows:
            raise ValueError('number of values must match the number of rows in flux')

        if name not in self.headers:
            self.append_columns(name, values=values)
        else:
            i = self.headers.get(str(name), name)
            for row, v in zip(self, values):
                row.values[i] = v

    def __iter__(self):
        return (row for row in self.matrix[1:])

    def __add__(self, rows):
        flux_final = self.copy()
        flux_final.append_rows(rows)

        return flux_final

    def __iadd__(self, rows):
        self.append_rows(rows)

        return self

    def __repr__(self):
        if self.is_empty:
            return '(empty)'

        headers = list(self.headers.keys())
        headers = str(headers).replace("'", '').replace('"', '')

        return '({:,})  {}'.format(self.num_rows, headers)

    # region {validation functions}
    @staticmethod
    def __validate_modify_matrix(m):
        """ validate matrix datatype and iteration depth; convert matrix to list of lists if neccessary

        if rows are tuples, they should prob be converted to lists, but not going to check here.
        Just let flux methods fail, user will be smarter when passing matrix next time

        why tf should primitive values be extracted, if these are already flux_row_cls / lev_row_cls?
            faster
            allow lev_row_cls.address info to be preserved
            if is_vengeance_class(m):
                list(m.flux_rows())

            if is_flux_row_class(m[0]):
                check header row, return m
        """
        if m is None:
            return [[]]

        if is_vengeance_class(m):
            return list(m.rows())

        base_names = set(base_class_names(m))

        if 'DataFrame' in base_names:
            raise NotImplementedError('DataFrame not supported')
        elif 'ndarray' in base_names:
            raise NotImplementedError('ndarray not supported (must be native python lists)')
        elif is_exhaustable(m):
            raise NotImplementedError('exhaustable iterators must be converted to lists')
        elif not is_subscriptable(m):
            raise TypeError("invalid type: <'{}'> (matrix should be a list of lists)".format(type(m)))
        elif iteration_depth(m) < 2:
            raise IndexError('matrix must have at least two dimensions (one dimension for rows, one for columns)')
        elif is_flux_row_class(m[0]):

            first_row = m[0]
            if first_row.is_header_row():
                m = [row.values for row in m]
            else:
                m = [first_row.names] + \
                    [row.values for row in m]

        return m

    @classmethod
    def __validate_modify_inserted_names(cls, inserted, headers):
        inserted = modify_iteration_depth(inserted, depth=0)

        nd = iteration_depth(inserted)
        if nd == 1 and len(inserted) == 2:
            inserted = [inserted]
        elif nd != 2:
            raise IndexError('inserted values should be tuples')

        before, names = [n for n in zip(*inserted)]
        cls.__validate_no_index_errors_with_headers(before, headers)

        cls.__validate_no_duplicates(names)
        cls.__validate_no_overlap_with_headers(names, headers)

        return inserted

    @staticmethod
    def __validate_no_reserved_headers(names):
        conflicting = set(names) & flux_row_cls.class_names
        if conflicting:
            raise ValueError('conflicting name(s) {} found in header row: {}'.format(list(conflicting), names))

    @staticmethod
    def __validate_no_duplicates(names):
        duplicates = [n for n, count in Counter(names).items() if count > 1]
        if duplicates:
            raise ValueError('duplicate column name(s) detected:\n{}'.format(duplicates))

    @staticmethod
    def __validate_no_overlap_with_headers(names, headers):
        invalid = list(headers.keys() & set(names))
        if invalid:
            raise ValueError('column name(s) conflict with existing headers:\n{}'.format(invalid))

    @staticmethod
    def __validate_complete_overlap_with_headers(names, headers):
        _names_ = set(s for s in names if not isinstance(s, int))
        invalid = _names_ - headers.keys()
        if invalid:
            raise ValueError('column(s) do not exist:\n{}'.format(invalid))

    @classmethod
    def __validate_no_index_errors_with_headers(cls, names, headers):
        cls.__validate_complete_overlap_with_headers(names, headers)

        num_cols = len(headers) - 1
        invalid  = [i for i in names if isinstance(i, int) if i > num_cols]
        if invalid:
            raise IndexError('column(s) out of bounds:\n{}'.format(invalid))

    @staticmethod
    def __names_to_column_indices(names, headers):
        _indices_ = list(range(len(headers)))

        indices = [headers.get(str(h), h) for h in names]       # convert to integers
        indices = [_indices_[i] for i in indices]               # handle negative indices

        return indices

    @staticmethod
    def __validate_headers_as_namedtuple_fields(headers):
        names = [n for n in headers.keys() if re.search('^[^a-z]|[ ]', str(n), re.I)]
        if names:
            raise ValueError('invalid field(s) for namedtuple constructor: {}'.format(names))

    # endregion






import _pickle as cpickle

import csv
import json
import os
import shutil

from datetime import datetime
from glob import glob

from .text import p_json_dumps
from .text import repr_


def read_file(path, encoding=None, mode='r'):

    extn = file_extension(path, include_dot=True)
    __validate_extension(extn)

    if extn == '.csv':
        with open(path, mode, encoding=encoding) as f:
            return list(csv.reader(f))

    if extn == '.json':
        with open(path, mode, encoding=encoding) as f:
            return json.load(f)

    if extn in {'.flux', '.pkl', '.pickle'}:
        if not mode.endswith('b'):
            mode += 'b'
        with open(path, mode) as f:
            return cpickle.load(f)

    with open(path, mode, encoding=encoding) as f:
        return f.read()


def write_file(path, data, encoding=None, mode='w'):

    extn = file_extension(path, include_dot=True)
    __validate_extension(extn)

    if extn == '.csv':
        with open(path, mode, encoding=encoding) as f:
            csv.writer(f, lineterminator='\n').writerows(data)

        return

    if extn == '.json':
        if not isinstance(data, str):
            data = p_json_dumps(data, ensure_ascii=(encoding is None))
        with open(path, mode, encoding=encoding) as f:
            f.write(data)

        return

    if extn in {'.flux', '.pkl', '.pickle'}:
        if not mode.endswith('b'):
            mode += 'b'
        with open(path, mode) as f:
            cpickle.dump(data, f)

        return

    # f.write() is faster than f.writelines(): convert data to string
    if not isinstance(data, str):
        data = repr_(data,
                     concat='\n',
                     quotes=False,
                     wrap=False)

    with open(path, mode, encoding=encoding) as f:
        f.write(data)


def __validate_extension(extn):
    """ check for file types that require specialized io libraries / protocols """
    if extn.startswith('.xls') or extn in {'.7z', '.gzip', '.hd5'}:
        raise NotImplementedError("'{}' file type not supported".format(extn))


def clear_dir(filedir):
    if not os.path.exists(filedir):
        return

    filedir = standardize_dir(filedir)

    for item in os.listdir(filedir):
        path = filedir + item
        if os.path.isdir(path):
            shutil.rmtree(path)
        else:
            os.remove(path)


def copy_dir(s_dir, d_dir, exclude_dirs=None):
    exclude_dirs = exclude_dirs or set()

    s_dir = standardize_dir(s_dir)
    d_dir = standardize_dir(d_dir)

    if not os.path.exists(d_dir):
        os.makedirs(d_dir)

    for file_content in os.listdir(s_dir):
        if file_content in exclude_dirs:
            continue

        s_path = s_dir + file_content
        d_path = d_dir + file_content

        if os.path.isdir(s_path):
            shutil.copytree(src=s_path, dst=d_path)
        else:
            shutil.copy(src=s_path, dst=d_path)


def file_creation_date(path):
    unix_t = os.path.getctime(path)
    return datetime.fromtimestamp(unix_t)


def file_last_modified(path):
    unix_t = os.stat(path).st_mtime
    return datetime.fromtimestamp(unix_t)


def standardize_dir(filedir, pathsep='/'):
    """
    pathsep=os.path.sep?
    """
    filedir = (filedir.replace('\\', pathsep)
                      .replace('/', pathsep)
                      .lower()
                      .strip())

    if filedir != '' and not filedir.endswith(pathsep):
        filedir += pathsep

    return filedir


def standardize_file_name(filename):
    return filename.lower().strip()


def standardize_path(path):
    filedir, filename = parse_path(path)

    filedir  = standardize_dir(filedir)
    filename = standardize_file_name(filename)

    return filedir + filename


def sanatize_file_name(filename):
    """ replace illegal Windows file name characters with '-'

    (there's an additional set of path characters that are
     illegal only for Windows' .zip compression)
    """
    invalid_chrs = {'\\': '-',
                    '/':  '-',
                    ':':  '-',
                    '*':  '-',
                    '?':  '-',
                    '<':  '-',
                    '>':  '-',
                    '|':  '-',
                    '"':  '-'}

    for k, v in invalid_chrs.items():
        filename = filename.replace(k, v)

    return filename


def assert_path_exists(path):
    filedir, filename = parse_path(path)

    if filedir != '' and not os.path.exists(filedir):
        raise FileNotFoundError("invalid directory: '{}'".format(filedir))

    if not os.path.exists(path):
        extn  = file_extension(filename, include_dot=True)
        retry = filename.replace(extn, '.*')
        path  = glob(standardize_dir(filedir) + retry)

        if path:
            msg = "invalid file extension".format(extn)
        else:
            msg = "'{}' not found within directory '{}'".format(filename, filedir)

        raise FileNotFoundError(msg)


def parse_path(path):

    if os.path.isdir(path):
        filedir  = path
        filename = ''
    else:
        filedir, filename = os.path.split(path)

    return filedir, filename


def file_extension(filename, include_dot=True):
    _, extn = os.path.splitext(filename)
    if include_dot is False:
        extn = extn.replace('.', '')

    extn = extn.lower().strip()

    return extn


def apply_file_extension(filename, extn):
    if not extn.startswith('.'):
        extn = '.' + extn

    if filename.endswith(extn):
        return filename

    return filename + extn






from collections import OrderedDict
from collections import namedtuple
from ..util.iter import map_numeric_indices


# noinspection DuplicatedCode
class flux_row_cls:

    class_names = {'_headers',
                   '_view_as_array',
                   'names',
                   'values',
                   'is_header_row',
                   'dict',
                   'namedtuples'}

    def __init__(self, headers, values):
        """
        :param headers: OrderedDict of {'header': index}
            headers is a single dictionary passed byref from the flux_cls to many flux_row_cls instances
            this eliminates need for all flux_row_cls objects to maintain a seprate copy of these mappings and
            allows for centralized and instantaneous updatdes
        :param values: list of underlying data

        (properties must be set on self.__dict__ instead of directly on self to prevent
         premature __setattr__ lookups)
        """
        self.__dict__['_headers'] = headers
        self.__dict__['values']   = values

    @property
    def _view_as_array(self):
        """ to help with debugging; meant to trigger a debugging feature in PyCharm

        PyCharm will recognize returned the (name, value) pairs as an ndarray
        and enable the "...view as array" option in the debugger which displays
        values in a special window as a table
        """
        import numpy
        return numpy.transpose([self.names, self.values])

    @property
    def names(self):
        return list(self._headers.keys())

    def is_header_row(self):
        """ determine if underlying values match self._headers.keys

        self.names == self.values will not always work, since map_numeric_indices()
        was used to modify self._headers values into more suitable dictionary keys,
        like modifying duplicate values to ensure they are unique, etc
        """
        header_names = list(map_numeric_indices(self.values).keys())
        return self.names == header_names

    def dict(self):
        return OrderedDict(zip(self.names, self.values))

    def namedtuples(self):
        return namedtuple('flux_row_ntc', self.names)(*self.values)

    def __getattr__(self, name):
        """  eg:
             v = row.column
        """
        try:
            i = self._headers.get(name, name)
            return self.values[i]
        except (TypeError, IndexError) as e:
            raise AttributeError(self.__attr_err_msg(name)) from e

    def __getitem__(self, name):
        """ eg:
            v = row['column']
        """
        try:
            i = self._headers.get(name, name)
            return self.values[i]
        except (TypeError, IndexError) as e:
            raise AttributeError(self.__attr_err_msg(name)) from e

    def __setattr__(self, name, value):
        """ eg:
            row.column = v
        """
        try:
            i = self._headers.get(name, name)
            self.values[i] = value
        except (TypeError, IndexError) as e:
            if name in self.__dict__:
                self.__dict__[name] = value
            else:
                raise AttributeError(self.__attr_err_msg(name)) from e

    def __setitem__(self, name, value):
        """ eg:
            row['column'] = v
        """
        try:
            i = self._headers.get(name, name)
            self.values[i] = value
        except (TypeError, IndexError) as e:
            if name in self.__dict__:
                self.__dict__[name] = value
            else:
                raise AttributeError(self.__attr_err_msg(name)) from e

    def __attr_err_msg(self, name):
        if isinstance(name, slice):
            return 'slice should be used directly on row.values\n(eg, row.values[2:5], not row[2:5])'

        names = '\n\t'.join(str(n) for n in self.names)
        return "No flux_row_cls column named '{}'\navailable columns:\n\t{}".format(name, names)

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, d):
        super().__setattr__('__dict__', d)

    def __len__(self):
        return len(self.values)

    def __bool__(self):
        return bool(self.values)

    def __iter__(self):
        return iter(self.values)

    def __eq__(self, other):
        return hash(self) == hash(other)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return id(self._headers) + hash(tuple(self.values))

    def __repr__(self):
        i = self.__dict__.get('i', '')
        if i != '':
            i = '({:,})  '.format(i)

        return '{}{}'.format(i, repr(self.values))


class lev_row_cls(flux_row_cls):

    class_names = {'_headers',
                   '_view_as_array',
                   'names',
                   'values',
                   'address',
                   'is_header_row',
                   'dict',
                   'namedtuples'}

    def __init__(self, headers, values, address=''):
        super().__init__(headers, values)

        self.__dict__['address'] = address

    def __repr__(self):
        return '{} {}'.format(self.address, repr(self.values))





from functools import lru_cache

from datetime import date
from datetime import datetime
from datetime import timedelta

from dateutil.parser import parse as dateutil_parse


def to_datetime(v, d_format=None):
    """
    :param v:
        value to be converted
    :param d_format:
        datetime.strptime format
    """
    if isinstance(v, str) or d_format is not None:
        date_time = parse_date_string(v, d_format)
    elif isinstance(v, float):
        date_time = parse_date_float(v) or parse_date_excel_serial(v)
    elif isinstance(v, datetime):
        date_time = v
    elif type(v) == date:
        date_time = datetime(v.year, v.month, v.day)
    elif isinstance(v, (list, tuple)):
        date_time = [to_datetime(d, d_format) for d in v]
    else:
        raise ValueError("dont know how to convert '{}' instance to datetime".format(type(v)))

    return date_time


def is_datetime(v):
    """ :retutn: boolean if converted, converted value
    """
    try:
        return True, to_datetime(v)
    except ValueError:
        return False, v


@lru_cache(maxsize=2**13)
def parse_date_float(f):
    s = str(int(f))

    try:
        y, m, d = (int(s[:4]),
                   int(s[4:6]),
                   int(s[6:8]))

        return datetime(y, m, d)
    except ValueError:
        return None


@lru_cache(maxsize=2**13)
def parse_date_excel_serial(f):
    """ number of days since 1900-01-01 """
    excel_epoch = datetime(1900, 1, 1)
    return excel_epoch + timedelta(days=int(f))


@lru_cache(maxsize=2**13)
def parse_date_string(s, d_format):
    s = s.strip()

    if d_format is not None:
        return datetime.strptime(s, d_format)

    date_time = __parse_date_strptime(s)
    if date_time is None:
        date_time = dateutil_parse(s)

    return date_time


def __parse_date_strptime(s):
    common_formats = ('%m-%d-%Y',  # 01-01-2000
                      '%m/%d/%Y',  # 01/01/2000
                      '%Y-%m-%d',  # 2000-01-01
                      '%Y/%m/%d',  # 2000/01/01
                      '%Y-%b-%d',  # 2000-jan-01
                      '%d-%b-%Y',  # 01-jan-2000
                      '%Y%m%d')    # 20000101

    if not (__compatible_with_strptime_lengths(s, common_formats)):
        return None

    for d_format in common_formats:
        try:
            return datetime.strptime(s, d_format)
        except ValueError:
            pass

    return None


def __compatible_with_strptime_lengths(s, common_formats):
    _common_formats_ = [df.replace('%Y', '2000')
                          .replace('%m', '01')
                          .replace('%d', '01')
                          .replace('%b', 'jan') for df in common_formats]

    max_cf = max(len(cf) for cf in _common_formats_)
    min_cf = min(len(cf) for cf in _common_formats_) - 2
    len_s = len(s)

    return min_cf <= len_s <= max_cf





from .dates import to_datetime
from .dates import is_datetime

from .filesystem import read_file
from .filesystem import write_file

from .iter import divide_sequence

from .text import print_runtime


__all__ = ['to_datetime',
           'is_datetime',
           'read_file',
           'write_file',
           'divide_sequence',
           'print_runtime']




import ctypes
import os
import pythoncom

from ctypes import POINTER
from ctypes import PyDLL
from ctypes import byref
from ctypes import c_void_p
from ctypes.wintypes import BOOL

from comtypes import COMError
from comtypes import GUID
from comtypes import IUnknown
from comtypes.automation import IDispatch as comtypes_idispatch
from comtypes.client import CreateObject as comtypes_createobject
from comtypes.client.dynamic import Dispatch as comtypes_dispatch

# noinspection PyUnresolvedReferences
from pythoncom import com_error as pythoncom_error
from win32com.client.gencache import EnsureDispatch

from ..util.filesystem import assert_path_exists
from ..util.filesystem import standardize_path
from ..util.text import vengeance_message
from .excel_constants import xlMaximized

# FindWindowExA only accepts ascii-encoded strings
FindWindowExA              = ctypes.windll.user32.FindWindowExA
SetForegroundWindow        = ctypes.windll.user32.SetForegroundWindow
AccessibleObjectFromWindow = ctypes.oledll.oleacc.AccessibleObjectFromWindow

corrupt_hwnds = set()

# FindWindowExA only accepts ascii-encoded strings
xl_main_ascii = 'XLMAIN'.encode('ascii')
xl_desk_ascii = 'XLDESK'.encode('ascii')
excel7_ascii  = 'EXCEL7'.encode('ascii')

xl_clsid  = '{00020400-0000-0000-C000-000000000046}'
native_om = -16


def open_workbook(path,
                  excel_app='new',
                  *,
                  read_only=False,
                  update_links=True):

    wb = is_workbook_open(path)
    if wb is None:
        wb = __open_workbook(excel_app, path, update_links, read_only)

    if (wb.ReadOnly is False) and read_only:
        raise AssertionError("'{}' is not opened read-only".format(wb.Name))

    if wb.ReadOnly and (read_only is False):
        vengeance_message("('{}' opened as read-only)".format(wb.Name))

    return wb


def is_workbook_open(path):
    """ scan all open workbooks across all Excel instances, match is determined by matching file paths
    :return captured Workbook or None
    """
    path = os.path.realpath(path)
    path = standardize_path(path)
    assert_path_exists(path)

    window_h = FindWindowExA(0, 0, xl_main_ascii, None)

    while window_h != 0:
        excel_app = __excel_application_from_hwnd(window_h)

        if excel_app:
            for wb in excel_app.Workbooks:
                if standardize_path(wb.FullName) == path:
                    return wb

        window_h = FindWindowExA(0, window_h, xl_main_ascii, None)

    return None


# noinspection PyUnusedLocal
def close_workbook(wb, save):
    """
    provides a few conveniences that the client may not want to deal with
        1) if Workbook is the only one in the Excel application, closes the application as well
        2) correctly releases the com pointer for the workbook
    """
    if save and wb.ReadOnly:
        raise AssertionError("'{}' is open as read-only, cannot save and close".format(wb.Name))

    excel_app = wb.Application
    display_alerts = excel_app.DisplayAlerts

    if save:
        wb.Save()

    excel_app.DisplayAlerts = False

    wb.Close()
    wb = None

    if excel_app.Workbooks.Count == 0:
        excel_app.Quit()
        excel_app = None
    else:
        excel_app.DisplayAlerts = display_alerts


def new_excel_instance():
    excel_app = comtypes_createobject('Excel.Application', dynamic=True)
    excel_app = __comtype_pointer_to_pythoncom(excel_app, comtypes_idispatch)
    excel_app = EnsureDispatch(excel_app)

    excel_app.WindowState = xlMaximized

    # excel_app_to_foreground(excel_app)
    reload_all_add_ins(excel_app)

    return excel_app


def empty_excel_instance():
    window_h = FindWindowExA(0, 0, xl_main_ascii, None)

    while window_h != 0:
        excel_app = __excel_application_from_hwnd(window_h)

        if __is_excel_app_empty(excel_app):
            vengeance_message('utilizing empty Excel instance')
            return excel_app

        window_h = FindWindowExA(0, window_h, xl_main_ascii, None)

    raise AssertionError('empty Excel instance not found')


def any_excel_instance():
    excel_app = (EnsureDispatch('Excel.Application')
                 or new_excel_instance())
    excel_app.Visible = True

    return excel_app


def all_excel_instances():
    excel_apps = []
    window_h = FindWindowExA(0, 0, xl_main_ascii, None)

    while window_h != 0:
        excel_app = __excel_application_from_hwnd(window_h)
        if excel_app:
            excel_apps.append(excel_app)

        window_h = FindWindowExA(0, window_h, xl_main_ascii, None)

    return excel_apps


def reload_all_add_ins(excel_app):
    vengeance_message('reloading Excel add-ins...')

    for add_in in excel_app.AddIns:
        if add_in.Installed:
            name = add_in.Name
            try:
                add_in.Installed = False
                add_in.Installed = True
                print('\t   * {}'.format(name))
            except COMError:
                vengeance_message('failed to load add-in: {}' + name)


def excel_application_to_foreground(excel_app):
    excel_app.Visible = True
    SetForegroundWindow(excel_app.Hwnd)


def __is_excel_app_empty(excel_app):
    if excel_app is None:
        return False

    workbooks = list(excel_app.Workbooks)
    if not workbooks:
        excel_app.Visible = True
        return True

    if len(workbooks) == 1:
        wb = workbooks[0]

        if wb.Saved and wb.Name == 'Book1':
            ws = wb.Sheets[1]
            if ws.Name == 'Sheet1' and ws.UsedRange.Address == '$A$1':
                return True

    return False


def __open_workbook(excel_app, path, update_links, read_only):
    if read_only:
        vengeance_message('opening workbook as read-only ...')

    excel_app = __convert_excel_application(excel_app)

    excel_app.DisplayAlerts = False
    wb = excel_app.Workbooks.Open(path, update_links, read_only)
    excel_app.DisplayAlerts = True

    return wb


def __convert_excel_application(excel_app):
    if excel_app in (None, 'new'):
        excel_app = new_excel_instance()
        excel_application_to_foreground(excel_app)

        return excel_app

    if excel_app == 'empty':
        excel_app = empty_excel_instance()
        excel_application_to_foreground(excel_app)

        return excel_app

    if excel_app == 'any':
        return any_excel_instance()

    if not hasattr(excel_app, 'Workbooks'):
        raise TypeError("excel_app should be an application pointer or in (None, 'new', 'empty', 'any')")

    return excel_app


# noinspection PyTypeChecker
def __excel_application_from_hwnd(window_h):
    """
    comtypes library is used to search windows handles for Excel application,
    then the application pointer is converted to a pywin object
    """
    global corrupt_hwnds

    if window_h in corrupt_hwnds:
        return None

    desk_hwnd   = FindWindowExA(window_h, None, xl_desk_ascii, None)
    excel7_hwnd = FindWindowExA(desk_hwnd, None, excel7_ascii, None)

    if excel7_hwnd == 0:
        corrupt_hwnds.add(window_h)
        return None

    obj_ptr = POINTER(comtypes_idispatch)()
    xl_guid = GUID.from_progid(xl_clsid)

    AccessibleObjectFromWindow(excel7_hwnd,
                               native_om,
                               byref(xl_guid),
                               byref(obj_ptr))
    try:
        com_ptr   = comtypes_dispatch(obj_ptr).Application
        excel_app = __comtype_pointer_to_pythoncom(com_ptr, comtypes_idispatch)
    except (COMError, pythoncom_error, NameError) as e:
        raise ChildProcessError('remote procedure call to Excel application rejected\n\n'
                                '(check if cursor is still active within a cell somewhere, '
                                'Excel will reject automation calls while waiting on '
                                'user input)') from e

    try:
        excel_app = EnsureDispatch(excel_app)
        excel_app.Visible = True

        return excel_app
    except AttributeError as e:
        raise AttributeError('error dispatching Excel application from win32com module: \n'
                             'Deleting the win32com folder, then re-running function may resolve the error\n'
                             'gencache folder location: %userprofile%\\AppData\\Local\\Temp\\gen_py') from e


# noinspection PyTypeChecker
def __comtype_pointer_to_pythoncom(ptr, interface):
    com_obj = PyDLL(pythoncom.__file__).PyCom_PyObjectFromIUnknown

    com_obj.restype  = ctypes.py_object
    com_obj.argtypes = (POINTER(IUnknown), c_void_p, BOOL)

    # noinspection PyProtectedMember
    return com_obj(ptr._comobj,
                   byref(interface._iid_),
                   True)









from collections import OrderedDict
from collections import defaultdict
from collections import Iterable


def modify_iteration_depth(v, depth=0):
    """
    eg:
        'a'           = modify_iteration_depth([['a']], depth=0)
        [2, 2, 2]     = modify_iteration_depth([[2, 2, 2]], depth=0)
        [[[1, 1, 1]]] = modify_iteration_depth([1, 1, 1], depth=3)
    """
    # v  = iterator_to_list(v, recurse=False)
    if is_exhaustable(v):
        raise TypeError('cannot mofify iteration depth of an exhaustable iterator')

    nd = iteration_depth(v, recurse=False)

    if nd > depth:
        for _ in range(nd - depth):
            if is_subscriptable(v) and len(v) == 1: v = v[0]
    elif nd < depth:
        for _ in range(depth - nd):                 v = [v]

    return v


def iterator_to_list(v, recurse=False):
    if not is_iterable(v):
        return v

    if is_vengeance_class(v):
        v = list(v.rows())
    elif not isinstance(v, (list, tuple)):
        v = list(v)

    if recurse:
        for i in range(len(v)):
            v[i] = iterator_to_list(v[i], recurse=True)

    return v


def iteration_depth(v, recurse=False):
    """
    eg:
        0 = iteration_depth('abc')
        1 = iteration_depth(['abc'])
        1 = iteration_depth([])
        2 = iteration_depth([[2], [2]])

    eg recurse = True:
        5 = iteration_depth([[2, 2, [3, 3, 3, [4, [5, 5, 5, 5], 4]]]], True)
    eg recurse = False:
        (only evaluates the depth of first elements)
        2 = iteration_depth([[2, 2, [3, 3, 3, [4, [5, 5, 5, 5], 4]]]], False)
    """
    if is_exhaustable(v):
        raise TypeError('cannot evaluate iteration depth of an exhaustable iterator')

    if not is_iterable(v):
        return 0

    if len(v) == 0:
        return 1

    if isinstance(v, dict):
        v = list(v.values())

    if recurse:
        return 1 + max(iteration_depth(_v_, recurse) for _v_ in v)

    try:
        return 1 + iteration_depth(v[0], recurse)
    except TypeError:
        return 1


def divide_sequence(sequence, divisions):
    """
    eg:
        [['a', 'a', 'a', 'a', 'a'],
         ['a', 'a', 'a', 'a', 'a']] = divide_sequence(['a'] * 10, 2)
    """
    num_items = len(sequence)
    if divisions > num_items:
        raise AssertionError('too many divisions')

    stride  = max(1, num_items // divisions)
    strides = [stride for _ in range(divisions)]
    undershoot = num_items - (stride * divisions)

    i = 0
    while undershoot > 0:
        strides[i] += 1
        undershoot -= 1
        i += 1

    i_1 = 0
    for stride in strides:
        i_2 = i_1 + stride
        yield sequence[i_1:i_2]
        i_1 = i_2


def is_iterable(v):
    """
    (strings are not considered to be iterable)

    eg:
        False = is_iterable('mike')
        True  = is_iterable(['m', 'i' 'k' 'e'])
    """
    is_iter   = isinstance(v, Iterable)
    is_string = isinstance(v, str)

    return is_iter and (not is_string)


def is_subscriptable(v):
    if isinstance(v, (list, tuple)):
        return True

    try:
        # noinspection PyStatementEffect
        v[0]
        return True

    except IndexError:
        return True
    except TypeError:
        return False


def is_exhaustable(v):
    return hasattr(v, '__next__')


def is_vengeance_class(o):
    """ all util functions should be independent from any other imports to avoid
    circular dependencies. Otherwise, isintance(o, (flux_cls, excel_levity_cls))
    would be sufficient
    """
    bases = set(base_class_names(o))
    return ('flux_cls' in bases) or ('excel_levity_cls' in bases)


def is_flux_row_class(o):
    bases = set(base_class_names(o))
    return 'flux_row_cls' in bases


def base_class_names(o):
    return [b.__name__ for b in o.__class__.mro()]


def transpose(m):
    if iteration_depth(m) == 1:
        t = [[row] for row in m]
    else:
        t = [list(row) for row in zip(*m)]

    return t


def map_numeric_indices(sequence, start=0):
    """ :return {value: index} for all items in sequence

    values are modified before they are added as keys:
        all values coerced to string
        '' are converted to 'None'
        non-unique keys are appended with '_{num nonunique}' suffix to make
        them unique and ensure that indices are incremented correctly

    eg
        {'a':   0,
         'b':   1,
         'c':   2,
         'b_2': 3} = map_numeric_indices(['a', 'b', 'c', 'b'])
    """
    indices   = OrderedDict()
    nonunique = defaultdict(int)

    for i, v in enumerate(sequence, start):
        v_s = str(v)

        if v_s == '':
            v_s = 'None'

        if v in nonunique:
            v_s = '{}_{}'.format(v_s, nonunique[v]+1)

        indices[v_s] = i
        nonunique[v] += 1

    return indices


def ordered_unique(sequence):
    d = OrderedDict((k, None) for k in sequence)
    return list(d.keys())


def invert_mapping(d):
    """
    return new mapping of
    {value: key}
        from
    {key: value}

    eg:
        {'a': [-1, -1, -1]
         'b': 2
         'c': 3
         'd': 3}
                to
        {-1: ['a', 'a', 'a']
         2:  b                      # single items not in list
         3:  ['c', 'd']}
    """
    if not isinstance(d, dict):
        raise TypeError('value must be a dictionary')

    inverted = OrderedDefaultDict(list)
    for k, v in d.items():
        _k_, _v_ = v, k

        for _k_ in modify_iteration_depth(v, 1):
            inverted[_k_].append(_v_)

    inverted = OrderedDict(inverted)
    for k, v in inverted.items():
        if len(v) == 1:
            inverted[k] = v[0]

    return inverted


class OrderedDefaultDict(OrderedDict):
    def __init__(self, default_factory=None,
                       *args,
                       **kwargs):

        if default_factory and (not callable(default_factory)):
            raise TypeError('first argument must be callable')

        self.default_factory = default_factory

        if args and (self.default_factory is list):
            self.__append_values(args[0])
        else:
            OrderedDict.__init__(self, *args, **kwargs)

    def __append_values(self, sequence):
        """ append all values as list if constructed default_factory is list """

        if isinstance(sequence, dict):
            sequence = OrderedDict(sequence.items()).items()
        elif not isinstance(sequence, (list, tuple)):
            raise ValueError

        for item in sequence:
            if len(item) == 2:
                self[item[0]].append(item[1])
            elif len(item) < 2:
                raise ValueError('need more than 1 value to unpack')
            elif len(item) > 2:
                raise ValueError('too many values to unpack (expected 2)')

    def __getitem__(self, key):
        try:
            return OrderedDict.__getitem__(self, key)
        except KeyError:
            return self.__missing__(key)

    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError(key)

        self[key] = value = self.default_factory()

        return value

    def defaultdict(self):
        return defaultdict(self.default_factory, self.items())

    def ordereddict(self):
        return OrderedDict(self)







import os
from textwrap import dedent

from logging import Logger
from logging import Formatter
from logging import FileHandler
from logging import StreamHandler
from logging import DEBUG


class log_cls(Logger):

    def __init__(self, name,
                       filedir=None,
                       log_format=None):

        super().__init__(name)

        self.log_format  = log_format
        self.formatter   = None
        self.child_desig = None

        self._callback       = None
        self._file_handler   = None
        self._stream_handler = None

        self.err_msg = ''

        self._set_level()
        self._add_formatter()
        self._add_file_handler(filedir)
        self._add_stream_handler()

    def print_message(self, msg):
        fh = self._file_handler
        sh = self._stream_handler

        if fh:
            fh.formatter = None
        if sh:
            sh.formatter = None

        self.info(msg)

        if fh:
            fh.formatter = self.formatter
        if sh:
            sh.formatter = self.formatter

    def add_parent(self, p_log):
        self.parent = p_log
        self._close_stream_handlers()

    def add_callback_function(self, f):
        self._callback = f

    def callback(self):
        if self._callback:
            self._callback()

    def _set_level(self):
        self.setLevel(DEBUG)

    def _add_formatter(self):
        if self.log_format is None:
            self.log_format = '%(asctime)s - %(levelname)s - %(message)s'

        self.formatter = Formatter(self.log_format)
        self.formatter.default_time_format = '%Y-%m-%d %I:%M:%S %p'

    def _add_file_handler(self, filedir):
        if filedir is None:
            return

        if not os.path.exists(filedir):
            os.makedirs(filedir)

        filename = self.name
        if not filename.endswith('.log'):
            filename += '.log'

        h = FileHandler(str(filedir) + filename, mode='w')
        h.setLevel(self.level)
        h.setFormatter(self.formatter)

        self.addHandler(h)
        self._file_handler = h

    def _add_stream_handler(self):
        h = StreamHandler()
        h.setLevel(self.level)
        h.setFormatter(self.formatter)

        self.addHandler(h)
        self._stream_handler = h

    def _close_stream_handlers(self):
        for h in self.__stream_handlers():
            h.close()
            self.removeHandler(h)

    def _close_file_handlers(self):
        for h in self.__file_handlers():
            h.close()
            self.removeHandler(h)

    def __stream_handlers(self):
        for h in self.handlers:
            if type(h) == StreamHandler:
                yield h

    def __file_handlers(self):
        for h in self.handlers:
            if isinstance(h, FileHandler):
                yield h

    def exception_handler(self, e_type, e_msg, e_trace):
        def frame_filename():
            return s_frame.tb_frame.f_code.co_filename

        if e_type and e_trace:
            self.err_msg = str(e_msg).replace('"', "'")

            e_type  = e_type.__name__
            c_frame = s_frame = e_trace
            has_child = not bool(self.child_desig)

            # naviagate to most recent stack frame
            while s_frame.tb_next is not None:
                if has_child is False:
                    if self.child_desig in frame_filename():
                        c_frame = s_frame
                        has_child = True

                s_frame = s_frame.tb_next

            exc_info = (e_type, e_msg, c_frame)
        else:
            self.err_msg = ''
            exc_info = None

        # include self.name?
        log_msg = dedent('''\n\n
        _______________________________________   vengeance  _____________________________________________
                          The result 'w+resign' was added to the game information
        
        <{}>
        "{}"
        _______________________________________   vengeance  _____________________________________________
        \n\n''').format(e_type, self.err_msg)

        # propagate error up through super class
        self.error(log_msg, exc_info=exc_info)

        self._close_file_handlers()
        self.callback()






from .excel_levity_cls import excel_levity_cls
from .flux_cls import flux_cls
from .flux_row_cls import flux_row_cls
from .log_cls import log_cls

__all__ = ['excel_levity_cls',
           'flux_cls',
           'flux_row_cls',
           'log_cls']




import os

import setuptools
from setuptools import setup

from textwrap import dedent


__version__ = '1.0.40'
__release__ = '$release 33'

long_description = 'https://github.com/michael-ross-ven/vengeance/blob/master/README.md\n(fill this out for pypi.org later)'


def write_version_file():
    f_dir = os.path.realpath(__file__)
    f_dir = os.path.split(f_dir)[0]

    with open(f_dir + '\\vengeance\\version.py', 'w') as f:
        s = '''
            # generated from setup.py
            __version__ = '{}'
            __release__ = '{}'
        '''.format(__version__, __release__)

        s = dedent(s) + '\n\n'

        f.write(s)


if __name__ == '__main__':
    write_version_file()

    setup(name='vengeance',
          version=__version__,
          description='Data library focusing on intuitive, row-major organization and interaction with the Excel object model',
          long_description=long_description,
          url='https://github.com/michael-ross-ven/vengeance',
          author='Michael Ross',
          author_email='michael.ross.uncc@gmail.com',
          license='MIT',
          install_requires=('comtypes', 'pypiwin32', 'python-dateutil'),
          # extra_require=('numpy', 'line-profiler'),
          packages=setuptools.find_packages(),
          classifiers=[
              "Programming Language :: Python :: 3",
              "License :: OSI Approved :: MIT License",
              "Operating System :: Microsoft :: Windows"
            ]

          )




try:
    from .version import __version__
    from .util import *
    from .classes import *
    from .excel_com import *

except ModuleNotFoundError as e:
    raise ModuleNotFoundError('\n\n\n*** vengeance import failure: {} ***\n'.format(e)) from e

except ImportError as e:
    raise ImportError('\n\n\n*** vengeance import failure: {} ***\n'.format(e)) from e


